@@ -24,8 +24,11 @@
 #include <locale.h>
 
 #define MAX_NUM_DRIVES 26
-#define FT70SEC 11644473600LL	       /* seconds between 1/1/1601 and
-				          1/1/1970 */
+#define FT70SEC 11644473600LL	       /* seconds between 1601-01-01 and
+				          1970-01-01 */
+#define FT80SEC 315529200	       /* seconds between 1970-01-01 and
+				          1980-01-01 */
+
 #define wsizeof(x) (sizeof(x)/sizeof(wchar_t))
 
 typedef struct _fdname {
@@ -84,6 +87,141 @@ static void remove_fdname(int fd)
     }
 }
 
+/* 
+ * The following UTF-8 validation is borrowed from
+ * ftp://ftp.unicode.org/Public/PROGRAMS/CVTUTF/ConvertUTF.c.
+ */
+
+/*
+ * Copyright 2001-2004 Unicode, Inc.
+ * 
+ * Disclaimer
+ * 
+ * This source code is provided as is by Unicode, Inc. No claims are
+ * made as to fitness for any particular purpose. No warranties of any
+ * kind are expressed or implied. The recipient agrees to determine
+ * applicability of information provided. If this file has been
+ * purchased on magnetic or optical media from Unicode, Inc., the
+ * sole remedy for any claim will be exchange of defective media
+ * within 90 days of receipt.
+ * 
+ * Limitations on Rights to Redistribute This Code
+ * 
+ * Unicode, Inc. hereby grants the right to freely use the information
+ * supplied in this file in the creation of products supporting the
+ * Unicode Standard, and to make copies of this file in any form
+ * for internal or external distribution as long as this notice
+ * remains attached.
+ */
+
+/*
+ * Index into the table below with the first byte of a UTF-8 sequence to
+ * get the number of trailing bytes that are supposed to follow it.
+ * Note that *legal* UTF-8 values can't have 4 or 5-bytes. The table is
+ * left as-is for anyone who may want to do such conversion, which was
+ * allowed in earlier algorithms.
+ */
+static const char trailingBytesForUTF8[256] = {
+    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
+    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
+    2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
+    3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5
+};
+
+/*
+ * Utility routine to tell whether a sequence of bytes is legal UTF-8.
+ * This must be called with the length pre-determined by the first byte.
+ * If not calling this from ConvertUTF8to*, then the length can be set by:
+ *  length = trailingBytesForUTF8[*source]+1;
+ * and the sequence is illegal right away if there aren't that many bytes
+ * available.
+ * If presented with a length > 4, this returns 0.  The Unicode
+ * definition of UTF-8 goes up to 4-byte sequences.
+ */
+
+static int isLegalUTF8(const unsigned char *source, int length)
+{
+    unsigned char a;
+    const unsigned char *srcptr = source + length;
+
+    switch (length) {
+	default:
+	    return 0;
+	    /* Everything else falls through when "1"... */
+	case 4:
+	    if ((a = (*--srcptr)) < 0x80 || a > 0xBF)
+		return 0;
+	case 3:
+	    if ((a = (*--srcptr)) < 0x80 || a > 0xBF)
+		return 0;
+	case 2:
+	    if ((a = (*--srcptr)) > 0xBF)
+		return 0;
+
+	    switch (*source) {
+		    /* no fall-through in this inner switch */
+		case 0xE0:
+		    if (a < 0xA0)
+			return 0;
+		    break;
+		case 0xED:
+		    if (a > 0x9F)
+			return 0;
+		    break;
+		case 0xF0:
+		    if (a < 0x90)
+			return 0;
+		    break;
+		case 0xF4:
+		    if (a > 0x8F)
+			return 0;
+		    break;
+		default:
+		    if (a < 0x80)
+			return 0;
+	    }
+
+	case 1:
+	    if (*source >= 0x80 && *source < 0xC2)
+		return 0;
+    }
+    if (*source > 0xF4)
+	return 0;
+    return 1;
+}
+
+/* End of code borrowed from ConvertUTF.c */
+
+int isLegalUTF8String(const unsigned char *source)
+{
+    const unsigned char *seq, *sourceend;
+    int seqlen;
+
+    sourceend = source + strlen(source);
+    seq = source;
+
+    while (seq < sourceend) {
+	seqlen = trailingBytesForUTF8[*seq] + 1;
+	if (!isLegalUTF8(seq, seqlen))
+	    return 0;
+	seq += seqlen;
+    }
+
+    return 1;
+}
+
 /* Translate an internal representation of a path (like /c/home) to
    a Windows path (like c:\home) */
 static wchar_t *intpath2winpath(const char *intpath)
@@ -95,6 +233,14 @@ static wchar_t *intpath2winpath(const char *intpath)
     wchar_t *lastslash;
     size_t intlen;
 
+    /* Verify that input is valid UTF-8. We cannot use MB_ERR_INVALID_CHARS
+       to MultiByteToWideChar, since it's only available in late versions of
+       Windows. */
+    if (!isLegalUTF8String(intpath)) {
+	logmsg(LOG_CRIT, "intpath2winpath: Illegal UTF-8 string:%s", intpath);
+	return NULL;
+    }
+
     /* Skip over multiple root slashes for paths like ///home/john */
     lastrootslash = intpath;
     while (*lastrootslash == '/')
@@ -113,7 +259,7 @@ static wchar_t *intpath2winpath(const char *intpath)
 
     if (!MultiByteToWideChar
 	(CP_UTF8, 0, lastrootslash, -1, winpath, winpath_len)) {
-	logmsg(LOG_CRIT, "intpath2winpath: MultiByteToWideChar failed\n");
+	logmsg(LOG_CRIT, "intpath2winpath: MultiByteToWideChar failed");
 	return NULL;
     }
 
@@ -194,6 +340,11 @@ int win_fchmod(int fildes, mode_t mode)
     int ret;
 
     winpath = intpath2winpath(get_fdname(fildes));
+    if (!winpath) {
+	errno = EINVAL;
+	return -1;
+    }
+
     ret = _wchmod(winpath, mode);
     free(winpath);
     return ret;
@@ -225,10 +376,33 @@ ssize_t pwrite(int fd, const void *buf, size_t count, off_t offset)
 {
     ssize_t size;
     off_t ret;
+    HANDLE h;
+    FILETIME ft;
+    SYSTEMTIME st;
+    ULARGE_INTEGER fti;
 
     if ((ret = lseek(fd, offset, SEEK_SET)) < 0)
 	return -1;
     size = write(fd, buf, count);
+
+    /* Since we are using the CreationTime attribute as "ctime", we need to
+       update it. From RFC1813: "Writing to the file changes the ctime in
+       addition to the mtime." */
+    h = (HANDLE) _get_osfhandle(fd);
+    GetSystemTime(&st);
+    SystemTimeToFileTime(&st, &ft);
+    /* Ceil up to nearest even second */
+    fti.LowPart = ft.dwLowDateTime;
+    fti.HighPart = ft.dwHighDateTime;
+    fti.QuadPart = ((fti.QuadPart + 20000000 - 1) / 20000000) * 20000000;
+    ft.dwLowDateTime = fti.LowPart;
+    ft.dwHighDateTime = fti.HighPart;
+    if (!SetFileTime(h, &ft, NULL, NULL)) {
+	fprintf(stderr,
+		"warning: pwrite: SetFileTime failed with error %ld\n",
+		GetLastError());
+    }
+
     return size;
 }
 
@@ -258,6 +432,7 @@ int win_init()
 	exit(1);
     }
 
+    /* init winsock */
     winsock_ver = MAKEWORD(1, 1);
     if (WSAStartup(winsock_ver, &wsadata)) {
 	fprintf(stderr, "Unable to initialise WinSock\n");
@@ -268,6 +443,10 @@ int win_init()
 	WSACleanup();
 	exit(1);
     }
+
+    /* disable error popups, for example from drives not ready */
+    SetErrorMode(SEM_FAILCRITICALERRORS);
+
     return 0;
 }
 
@@ -328,6 +507,10 @@ int win_stat(const char *file_name, backend_statstruct * buf)
     }
 
     winpath = intpath2winpath(file_name);
+    if (!winpath) {
+	errno = EINVAL;
+	return -1;
+    }
 
     ret = _wstat(winpath, &win_statbuf);
     if (ret < 0) {
@@ -404,6 +587,11 @@ int win_open(const char *pathname, int flags, ...)
     va_end(args);
 
     winpath = intpath2winpath(pathname);
+    if (!winpath) {
+	errno = EINVAL;
+	return -1;
+    }
+
     fd = _wopen(winpath, flags | O_BINARY, mode);
     free(winpath);
     if (fd < 0) {
@@ -447,6 +635,12 @@ UNFS3_WIN_DIR *win_opendir(const char *name)
 	ret->logdrives = GetLogicalDrives();
     } else {
 	winpath = intpath2winpath(name);
+	if (!winpath) {
+	    free(ret);
+	    errno = EINVAL;
+	    return NULL;
+	}
+
 	ret->stream = _wopendir(winpath);
 	free(winpath);
 	if (ret->stream == NULL) {
@@ -486,7 +680,7 @@ struct dirent *win_readdir(UNFS3_WIN_DIR * dir)
 	if (!WideCharToMultiByte
 	    (CP_UTF8, 0, de->d_name, -1, dir->de.d_name,
 	     sizeof(dir->de.d_name), NULL, NULL)) {
-	    logmsg(LOG_CRIT, "win_readdir: WideCharToMultiByte failed\n");
+	    logmsg(LOG_CRIT, "win_readdir: WideCharToMultiByte failed");
 	    return NULL;
 	}
 	return &dir->de;
@@ -531,6 +725,11 @@ int win_mkdir(const char *pathname, U(mode_t mode))
     }
 
     winpath = intpath2winpath(pathname);
+    if (!winpath) {
+	errno = EINVAL;
+	return -1;
+    }
+
     /* FIXME: Use mode */
     ret = _wmkdir(winpath);
     free(winpath);
@@ -584,6 +783,11 @@ int win_statvfs(const char *path, backend_statvfsstruct * buf)
     }
 
     winpath = intpath2winpath(path);
+    if (!winpath) {
+	errno = EINVAL;
+	return -1;
+    }
+
     winpath[3] = '\0';		       /* Cut off after x:\ */
 
     if (!GetDiskFreeSpaceW
@@ -622,6 +826,11 @@ int win_remove(const char *pathname)
     }
 
     winpath = intpath2winpath(pathname);
+    if (!winpath) {
+	errno = EINVAL;
+	return -1;
+    }
+
     ret = _wremove(winpath);
     free(winpath);
     return ret;
@@ -639,18 +848,29 @@ int win_chmod(const char *path, mode_t mode)
     }
 
     winpath = intpath2winpath(path);
+    if (!winpath) {
+	errno = EINVAL;
+	return -1;
+    }
+
     ret = _wchmod(winpath, mode);
     free(winpath);
     return ret;
 }
 
-int win_utime(const char *path, const struct utimbuf *times)
+/* 
+   If creation is false, the LastAccessTime will be set according to
+   times->actime. Otherwise, CreationTime will be set. LastWriteTime
+   is always set according to times->modtime.
+*/
+static int win_utime_creation(const char *path, const struct utimbuf *times,
+			      int creation)
 {
     wchar_t *winpath;
     int ret = 0;
     HANDLE h;
     ULARGE_INTEGER fti;
-    FILETIME atime, mtime;
+    FILETIME xtime, mtime;
 
     if (!strcmp("/", path)) {
 	/* Emulate root */
@@ -659,12 +879,16 @@ int win_utime(const char *path, const struct utimbuf *times)
     }
 
     winpath = intpath2winpath(path);
+    if (!winpath) {
+	errno = EINVAL;
+	return -1;
+    }
 
     /* Unfortunately, we cannot use utime(), since it doesn't support
        directories. */
     fti.QuadPart = UInt32x32To64(times->actime + FT70SEC, 10000000);
-    atime.dwHighDateTime = fti.HighPart;
-    atime.dwLowDateTime = fti.LowPart;
+    xtime.dwHighDateTime = fti.HighPart;
+    xtime.dwLowDateTime = fti.LowPart;
     fti.QuadPart = UInt32x32To64(times->modtime + FT70SEC, 10000000);
     mtime.dwHighDateTime = fti.HighPart;
     mtime.dwLowDateTime = fti.LowPart;
@@ -673,7 +897,8 @@ int win_utime(const char *path, const struct utimbuf *times)
 		    FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING,
 		    FILE_ATTRIBUTE_NORMAL | FILE_FLAG_BACKUP_SEMANTICS, NULL);
 
-    if (!SetFileTime(h, NULL, &atime, &mtime)) {
+    if (!SetFileTime
+	(h, creation ? &xtime : NULL, creation ? NULL : &xtime, &mtime)) {
 	errno = EACCES;
 	ret = -1;
     }
@@ -683,6 +908,11 @@ int win_utime(const char *path, const struct utimbuf *times)
     return ret;
 }
 
+int win_utime(const char *path, const struct utimbuf *times)
+{
+    return win_utime_creation(path, times, FALSE);
+}
+
 int win_rmdir(const char *path)
 {
     wchar_t *winpath;
@@ -695,6 +925,11 @@ int win_rmdir(const char *path)
     }
 
     winpath = intpath2winpath(path);
+    if (!winpath) {
+	errno = EINVAL;
+	return -1;
+    }
+
     ret = _wrmdir(winpath);
     free(winpath);
     return ret;
@@ -712,7 +947,16 @@ int win_rename(const char *oldpath, const char *newpath)
     }
 
     oldwinpath = intpath2winpath(oldpath);
+    if (!oldwinpath) {
+	errno = EINVAL;
+	return -1;
+    }
     newwinpath = intpath2winpath(newpath);
+    if (!newwinpath) {
+	free(oldwinpath);
+	errno = EINVAL;
+	return -1;
+    }
 
     ret = _wrename(oldwinpath, newwinpath);
     free(oldwinpath);
@@ -752,16 +996,26 @@ int win_utf8ncasecmp(const char *s1, const char *s2, size_t n)
     wchar_t ws1[4096], ws2[4096];
     int converted;
 
+    /* Make sure input is valid UTF-8 */
+    if (!isLegalUTF8String(s1)) {
+	logmsg(LOG_CRIT, "win_utf8ncasecmp: Illegal UTF-8 string:%s", s1);
+	return -1;
+    }
+    if (!isLegalUTF8String(s2)) {
+	logmsg(LOG_CRIT, "win_utf8ncasecmp: Illegal UTF-8 string:%s", s2);
+	return -1;
+    }
+
     /* Convert both strings to wide chars */
     converted = MultiByteToWideChar(CP_UTF8, 0, s1, n, ws1, wsizeof(ws1));
     if (!converted) {
-	logmsg(LOG_CRIT, "win_utf8ncasecmp: MultiByteToWideChar failed\n");
+	logmsg(LOG_CRIT, "win_utf8ncasecmp: MultiByteToWideChar failed");
 	return -1;
     }
     ws1[converted] = '\0';
     converted = MultiByteToWideChar(CP_UTF8, 0, s2, n, ws2, wsizeof(ws2));
     if (!converted) {
-	logmsg(LOG_CRIT, "win_utf8ncasecmp: MultiByteToWideChar failed\n");
+	logmsg(LOG_CRIT, "win_utf8ncasecmp: MultiByteToWideChar failed");
 	return 1;
     }
     ws2[converted] = '\0';
@@ -770,4 +1024,39 @@ int win_utf8ncasecmp(const char *s1, const char *s2, size_t n)
     return _wcsicmp(ws1, ws2);
 }
 
+static void win_verf_to_ubuf(struct utimbuf *ubuf, createverf3 verf)
+{
+    ubuf->actime = verf[0] | verf[1] << 8 | verf[2] << 16 | verf[3] << 24;
+    ubuf->modtime = verf[4] | verf[5] << 8 | verf[6] << 16 | verf[7] << 24;
+
+    /* FAT can only store dates in the interval 1980-01-01 to 2107-12-31.
+       However, since the utime interface uses Epoch time, we are further
+       limited to 1980-01-01 to 2038-01-19, assuming 32 bit signed time_t.
+       math.log(2**31-1 - FT80SEC, 2) = 30.7, which means that we can only
+       use 30 bits. */
+    ubuf->actime &= 0x3fffffff;
+    ubuf->actime += FT80SEC;
+    ubuf->modtime &= 0x3fffffff;
+    ubuf->modtime += FT80SEC;
+    /* While FAT CreationTime has a resolution of 10 ms, WriteTime only has a 
+       resolution of 2 seconds. */
+    ubuf->modtime &= ~1;
+}
+
+int win_store_create_verifier(char *obj, createverf3 verf)
+{
+    struct utimbuf ubuf;
+
+    win_verf_to_ubuf(&ubuf, verf);
+    return win_utime_creation(obj, &ubuf, TRUE);
+}
+
+int win_check_create_verifier(backend_statstruct * buf, createverf3 verf)
+{
+    struct utimbuf ubuf;
+
+    win_verf_to_ubuf(&ubuf, verf);
+    return (buf->st_ctime == ubuf.actime && buf->st_mtime == ubuf.modtime);
+}
+
 #endif				       /* WIN32 */
